# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import shutil
import subprocess
from typing import Optional


def which(cmd: str) -> Optional[str]:
    return shutil.which(cmd)


def is_tool(name: str) -> bool:
    return which(name) is not None


def run(app: str, *args: str, capture_output=False):
    cmd = which(app)
    if cmd is None:
        return None
    return subprocess.run(
        [cmd, *args], check=False, encoding="UTF-8", capture_output=capture_output
    )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import argparse
from dataclasses import dataclass, field
from typing import Any, List, Optional, Union


@dataclass
class Argument:
    help: str = ""
    pos: bool = False
    names: List[str] = field(default_factory=list)
    nargs: Union[str, int, None] = None
    meta: Optional[str] = None
    action: Union[str, argparse.Action, None] = None
    default: Optional[Any] = None
    choices: Optional[List[str]] = None


class FlagArgument(Argument):
    def __init__(self, help: str = "", names: List[str] = []):
        super().__init__(help=help, names=names, action="store_true", default=False)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

from cxx_flow.flow.step import SerialStep, register_step

from .store_packages import StorePackages
from .store_tests import StoreTests


class StoreBoth(SerialStep):
    name = "Store"

    def __init__(self):
        super().__init__()
        self.children = [StoreTests(), StorePackages()]


register_step(StoreBoth())



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import platform
from typing import Annotated

from ..flow.arg import Argument
from ..flow.config import Runtime
from ..flow.uname import uname


def command_system(
    format: Annotated[
        str,
        Argument(
            help="select, what format should be returned",
            choices=["props", "platform", "debug"],
        ),
    ],
    _: Runtime,
):
    """Produces system information for CI pipelines"""

    node = platform.node()
    system, version, arch = uname()

    if format == "props":
        print(f"-phost.name='{node}'")
        print(f"-pos='{system}'")
        if version is not None:
            print(f"-pos.version='{version}'")
        print(f"-parch='{arch}'")

    if format == "platform":
        version = "" if version is None else f"-{version}"
        print(f"{system}{version}-{arch}")

    if format == "debug":
        for name in [
            "uname",
            "machine",
            "node",
            "platform",
            "processor",
            "release",
            "system",
            "version",
        ]:
            print(name, platform.__dict__[name]())

        print("-----")
        print(f"node {node}")
        print(f"os {system}")
        if version is not None:
            print(f"version {version}")
        print(f"machine {arch}")



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import os

from cxx_flow.flow.config import Config, Runtime
from cxx_flow.flow.step import Step, register_step


class StoreTests(Step):
    name = "StoreTests"
    runs_after = ["Test"]

    def run(self, config: Config, rt: Runtime) -> int:
        return rt.cp(
            f"build/{config.preset}/test-results", "build/artifacts/test-results"
        )


register_step(StoreTests())



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)


import shutil
import sys
from typing import Annotated, List, Optional, Set, cast

from ..flow import dependency, matrix
from ..flow.arg import Argument
from ..flow.config import Configs, Runtime
from ..flow.step import Step


def command_run(
    steps: Annotated[
        Optional[List[str]],
        Argument(
            help="run only listed steps; if missing, run all the steps",
            names=["-s", "--steps"],
            nargs="*",
            meta="step",
            action="append",
            default=[],
        ),
    ],
    configs: Configs,
    rt: Runtime,
):
    """Runs automation steps for current project"""

    rt_steps = cast(List[Step], rt.steps)
    steps = matrix.flatten(step.split(",") for step in matrix.flatten(steps))
    if not steps:
        steps = [step.name for step in rt_steps]

    step_names = set(steps)
    program = [step for step in rt_steps if step.name in step_names]

    errors = gather_dependencies_for_all_configs(configs, rt, program)
    if len(errors) > 0:
        if not rt.silent:
            for error in errors:
                print(f"cxx-flow: {error}", file=sys.stderr)
        return 1

    printed = refresh_directories(configs, rt, program)
    return run_steps(configs, rt, program, printed)


def gather_dependencies_for_all_configs(
    configs: Configs, rt: Runtime, steps: List[Step]
):
    deps: List[dependency.Dependency] = []
    for config in configs.usable:
        active_steps = [step for step in steps if step.is_active(config, rt)]
        deps.extend(dependency.gather(active_steps))
    return dependency.verify(deps)


def refresh_directories(configs: Configs, rt: Runtime, steps: List[Step]):
    directories_to_refresh: Set[str] = set()
    for config in configs.usable:
        for step in steps:
            if step.is_active(config, rt):
                dirs = step.directories_to_remove(config)
                directories_to_refresh.update(dirs)

    printed = False
    for dirname in directories_to_refresh:
        if not rt.silent:
            printed = True
            print(f"[-] {dirname}")
        if not rt.dry_run:
            shutil.rmtree(dirname, ignore_errors=True)

    return printed


def run_steps(configs: Configs, rt: Runtime, program: List[Step], printed: bool) -> int:
    for config in configs.usable:
        steps = [step for step in program if step.is_active(config, rt)]
        step_count = len(steps)
        if step_count == 0:
            continue

        if printed:
            print()
        printed = True

        print("-", config.build_name)
        for index in range(step_count):
            step = steps[index]
            print(f"-- step {index + 1}/{step_count}: {step.name}")
            ret = step.run(config, rt)
            if ret:
                return 1

    return 0



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import importlib
import os
import sys
from dataclasses import dataclass
from types import ModuleType
from typing import List, Union

from . import step
from .config import FlowConfig, RunAlias


@dataclass
class Sorted:
    plugin: step.Step
    name: str
    runs_after: List[str]
    runs_before: List[str]

    @staticmethod
    def from_step(plugin: step.Step):
        name = plugin.name
        runs_after = [*plugin.runs_after]
        runs_before = [*plugin.runs_before]
        return Sorted(plugin, name, runs_after, runs_before)


def _load_plugins(directory: str, package: Union[str, None]):
    for _, dirnames, filenames in os.walk(directory):
        for dirname in dirnames:
            importlib.import_module(f".{dirname}", package=package)
        for filename in filenames:
            if filename == "__init__.py":
                continue
            importlib.import_module(
                f".{os.path.splitext(filename)[0]}", package=package
            )
        dirnames[:] = []
        pass


def _load_module_plugins(cfg: FlowConfig, mod: ModuleType):
    spec = mod.__spec__
    if not spec:
        return
    for location in spec.submodule_search_locations:
        _load_plugins(location, spec.name)


def _sort_steps():
    steps = step.__steps
    unsorted = [Sorted.from_step(step) for step in steps]
    known_names = [step.name for step in unsorted]

    for plugin in unsorted:
        for name in plugin.runs_before:
            for successor in unsorted:
                if successor.name != name:
                    continue
                successor.runs_after.append(plugin.name)

    for plugin in unsorted:
        runs_after: List[str] = []
        for name in plugin.runs_after:
            if name in known_names:
                runs_after.append(name)
        plugin.runs_after = runs_after

    result: List[step.Step] = []

    while len(unsorted) > 0:
        layer = [plugin for plugin in unsorted if len(plugin.runs_after) == 0]
        unsorted = [plugin for plugin in unsorted if len(plugin.runs_after) > 0]
        for plugin in layer:
            for remaining in unsorted:
                try:
                    remaining.runs_after.remove(plugin.name)
                except ValueError:
                    pass
        result.extend(plugin.plugin for plugin in layer)
        if len(layer) == 0:
            result.extend(plugin.plugin for plugin in unsorted)
            break

    return result


def load_steps(cfg: FlowConfig):
    std_plugins = importlib.import_module("cxx_flow.plugins")
    _load_module_plugins(cfg, std_plugins)

    local_plugins = os.path.abspath(os.path.join(".flow", "extensions"))
    if os.path.isdir(local_plugins):
        sys.path.append(local_plugins)
    for root, dirnames, _ in os.walk(local_plugins):
        for dirname in dirnames:
            init = os.path.join(root, dirname, "__init__.py")
            if not os.path.isfile(init):
                continue
            plugins = importlib.import_module(dirname)
            _load_module_plugins(cfg, plugins)
        dirnames[:] = []

    return _sort_steps()


def clean_aliases(cfg: FlowConfig, valid_steps: List[step.Step]):
    entries = cfg._cfg.get("entry")
    if not entries:
        return

    step_names = {step.name for step in valid_steps}

    keys_to_remove: List[str] = []
    for key in entries:
        entry = entries[key]
        steps = entry.get("steps")
        if not steps:
            keys_to_remove.append(key)
            continue
        rewritten: List[str] = []
        for step in steps:
            if step in step_names:
                rewritten.append(step)
        steps[:] = rewritten

        if not steps:
            keys_to_remove.append(key)

    for key in keys_to_remove:
        del entries[key]

    cfg.steps = valid_steps
    cfg.aliases = [RunAlias.from_json(key, entries[key]) for key in entries]
